import sys
from panther.test import APIClient
from unittest import TestCase


class TestSimpleRequests(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        sys.path.append('tests/app')
        from tests.app.main import app
        cls.client = APIClient(app=app)

    @classmethod
    def tearDownClass(cls) -> None:
        sys.path.pop()

    def test_header(self):
        res = self.client.get('request-header/')
        assert res.status_code == 200
        headers = [
            'accept_encoding', 'content_length', 'authorization', 'content_type', 'user_agent',
            'connection', 'accept', 'host', 'sec_websocket_version', 'sec_websocket_key', 'upgrade'
        ]
        assert [*res.data.keys()] == headers
        assert not all(res.data.values())

    def test_path(self):
        res = self.client.get('request-path/')
        assert res.status_code == 200
        assert res.data == '/request-path/'

    def test_client(self):
        res = self.client.get('request-client/')
        assert res.status_code == 200
        assert res.data == ['127.0.0.1', 8585]

    def test_query_params(self):
        res = self.client.get('request-query_params/', query_params={'my': 'name', 'is': 'ali', 'how': 'are'})
        assert res.status_code == 200
        assert res.data == {'my': 'name', 'is': 'ali', 'how': 'are'}



from panther.logger import logger

logo = r"""│    ____                 __    __                         │
│   /\  _`\              /\ \__/\ \                        │
│   \ \ \L\ \ __      ___\ \ ,_\ \ \___      __   _ __     │
│    \ \ ,__/'__`\  /' _ `\ \ \/\ \  _ `\  /'__`\/\`'__\   │
│     \ \ \/\ \L\.\_/\ \/\ \ \ \_\ \ \ \ \/\  __/\ \ \/    │
│      \ \_\ \__/.\_\ \_\ \_\ \__\\ \_\ \_\ \____\\ \_\    │
│       \/_/\/__/\/_/\/_/\/_/\/__/ \/_/\/_/\/____/ \/_/    │
"""

help_message = f"""╭{58 * '─'}╮
{logo}│{58 * ' '}│
│                                                          │
│   usage:                                                 │
│       - panther create <project_name> <directory>        │
│           Create project in <directory> default is `.`   │
│                                                          │
│       - panther run [--reload | --help]                  │
│           Run your project with uvicorn                  │
│                                                          │
│       - panther shell                                    │
│           Run interactive python shell                   │
│                                                          │
│       - panther monitor                                  │
│           Show the monitor :)                            │
│                                                          │
│       - panther version                                  │
│           Print the current version of Panther           │
│                                                          │
│       - panther [--help | -h ]                           │
│           Show this message and exit                     │
╰{58 * '─'}╯
"""

run_help_message = """Run `uvicorn --help` for more help"""


def cli_error(message: str | TypeError) -> None:
    logger.error(f'Error: {message}\n\nUse panther -h for more help')


def clean_args(args: list[str]) -> dict:
    """
    Input: ['--reload', '--host', '127.0.0.1', ...]
    Output: {'--reload: None, 'host': '127.0.0.1', ...}
    """
    _args = dict()
    for i, arg in enumerate(args):
        if arg.startswith('--'):
            if (i + 1) < len(args):
                _args[arg[2:]] = args[i + 1]
            else:
                _args[arg[2:]] = True
    return _args



import os
import sys
import platform

from rich import print as rprint

from panther import version as panther_version
from panther.cli.create_command import create
from panther.cli.monitor_command import monitor
from panther.cli.run_command import run
from panther.cli.utils import clean_args, cli_error, help_message


def shell() -> None:
    if platform.system().lower() == "windows":
        os.system('python')
    else:
        os.system('bpython')


def version() -> None:
    print(panther_version())


def start() -> None:
    if len(sys.argv) < 2:
        cli_error("Please pass some arguments to the command.")
    else:
        command = sys.argv and sys.argv[1] or None
        args = clean_args(sys.argv[2:])

        match command:
            case '-h' | '--help':
                rprint(help_message)
            case 'create':
                create(sys.argv[2:])
            case 'run':
                run(args)
            case 'shell':
                shell()
            case 'monitor':
                monitor()
            case 'version':
                version()
            case _:
                cli_error('Invalid Arguments.')



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

import orjson as json
from panther import status

from panther._utils import generate_ws_connection_id
from panther.base_request import BaseRequest
from panther.configs import config
from panther.logger import logger
from panther.utils import Singleton

if TYPE_CHECKING:
    from redis import Redis


class WebsocketConnections(Singleton):
    def __init__(self):
        self.connections = dict()
        self.connections_count = 0

    def __call__(self, r: Redis | None):
        if r:
            subscriber = r.pubsub()
            subscriber.subscribe('websocket_connections')
            logger.debug("Subscribing: 'websocket_connections'")
            for channel_data in subscriber.listen():
                logger.debug(f'{channel_data=}')
                # Check Type of PubSub Message
                match channel_data['type']:
                    case 'subscribe':
                        continue

                    case 'message':
                        loaded_data = json.loads(channel_data['data'].decode())
                        if (
                                isinstance(loaded_data, dict)
                                and 'connection_id' in loaded_data
                                and (data := loaded_data.get('data'))
                                and (action := loaded_data.get('action'))
                                and (connection := self.connections.get(connection_id := loaded_data['connection_id']))
                       ):
                            # Check Action of WS
                            match action:
                                case 'send':
                                    logger.debug(f'Sending Message to {connection_id}')
                                    asyncio.run(connection.send(data=data))
                                case 'close':
                                    try:
                                        asyncio.run(connection.close(code=data['code'], reason=data['reason']))
                                    except RuntimeError:
                                        # We are trying to disconnect the connection between a thread and a user
                                        # from another thread
                                        # it's working, but we have to find another solution for close the connection
                                        # Error:
                                        # Task <Task pending coro=<Websocket.close()>> got Future
                                        # <Task pending coro=<WebSocketCommonProtocol.transfer_data()>>
                                        # attached to a different loop
                                        pass
                                case _:
                                    logger.debug(f'Unknown Message Action: {action}')
                    case _:
                        logger.debug(f'Unknown Channel Type: {channel_data["type"]}')

    async def new_connection(self, connection: Websocket):
        await connection.connect(**connection.path_variables)
        if connection.is_connected:
            self.connections_count += 1

            # Save New ConnectionID
            self.connections[connection.connection_id] = connection

    def remove_connection(self, connection: Websocket):
        self.connections_count -= 1
        del self.connections[connection.connection_id]


class Websocket(BaseRequest):
    is_connected: bool = False

    async def connect(self, **kwargs):
        """
        Check your conditions then self.accept() the connection
        """
        await self.accept()

    async def accept(self, subprotocol: str = None, headers: dict = None):
        await self.asgi_send({'type': 'websocket.accept', 'subprotocol': subprotocol, 'headers': headers or {}})
        self.is_connected = True

        # Generate ConnectionID
        connection_id = generate_ws_connection_id()
        while connection_id in config['websocket_connections'].connections:
            connection_id = generate_ws_connection_id()

        # Set ConnectionID
        self.set_connection_id(connection_id)

    async def receive(self, data: str | bytes):
        pass

    async def send(self, data: any = None):
        if data:
            if isinstance(data, bytes):
                await self.send_bytes(bytes_data=data)
            elif isinstance(data, str):
                await self.send_text(text_data=data)
            else:
                await self.send_text(text_data=json.dumps(data).decode())

    async def send_text(self, text_data: str):
        await self.asgi_send({'type': 'websocket.send', 'text': text_data})

    async def send_bytes(self, bytes_data: bytes):
        await self.asgi_send({'type': 'websocket.send', 'bytes': bytes_data})

    async def close(self, code: int = status.WS_1000_NORMAL_CLOSURE, reason: str = ''):
        self.is_connected = False
        config['websocket_connections'].remove_connection(self)
        await self.asgi_send({'type': 'websocket.close', 'code': code, 'reason': reason})

    async def listen(self):
        while self.is_connected:
            response = await self.asgi_receive()
            if response['type'] == 'websocket.connect':
                continue

            if response['type'] == 'websocket.disconnect':
                break

            if 'text' in response:
                await self.receive(data=response['text'])
            else:
                await self.receive(data=response['bytes'])

    def set_path_variables(self, path_variables: dict):
        self._path_variables = path_variables

    @property
    def path_variables(self):
        return getattr(self, '_path_variables', {})

    def set_connection_id(self, connection_id):
        self._connection_id = connection_id

    @property
    def connection_id(self) -> str:
        connection_id = getattr(self, '_connection_id', None)
        if connection_id is None:
            logger.error('You should first `self.accept()` the connection then use the `self.connection_id`')
        return connection_id




from __future__ import annotations

from panther import status
from panther._utils import publish_to_ws_channel
from panther.base_websocket import Websocket, WebsocketConnections
from panther.configs import config
from panther.db.connection import redis


class GenericWebsocket(Websocket):

    async def connect(self, **kwargs):
        """
        Check your conditions then `accept()` the connection
        """
        await self.accept()

    async def receive(self, data: str | bytes):
        """
        Received `data` of connection,
        You may want to use json.loads() on the `data`
        """

    async def send(self, data: any = None):
        """
        We are using this method to send message to the client,
        You may want to override it with your custom scenario. (not recommended)
        """
        return await super().send(data=data)


async def send_message_to_websocket(connection_id: str, data: any):
    if redis.is_connected:
        publish_to_ws_channel(connection_id=connection_id, action='send', data=data)
    else:
        websocket_connections: WebsocketConnections = config['websocket_connections']
        if connection := websocket_connections.connections.get(connection_id):
            await connection.send(data=data)


async def close_websocket_connection(connection_id: str, code: int = status.WS_1000_NORMAL_CLOSURE, reason: str = ''):
    if redis.is_connected:
        data = {
            'code': code,
            'reason': reason
        }
        publish_to_ws_channel(connection_id=connection_id, action='close', data=data)
    else:
        websocket_connections: WebsocketConnections = config['websocket_connections']
        if connection := websocket_connections.connections.get(connection_id):
            await connection.close(code=code, reason=reason)



from pydantic import BaseModel


class File(BaseModel):
    file_name: str
    content_type: str
    file: bytes

    def __repr__(self) -> str:
        return f'{self.__repr_name__()}(file_name={self.file_name}, content_type={self.content_type})'

    __str__ = __repr__



import ast
import os
import platform
from datetime import timedelta
from importlib import import_module

from pydantic._internal._model_construction import ModelMetaclass

from panther._utils import import_class
from panther.cli.utils import cli_error
from panther.configs import JWTConfig, config
from panther.middlewares import BaseMiddleware
from panther.routings import finalize_urls, flatten_urls
from panther.throttling import Throttling

__all__ = (
    'load_configs_file',
    'load_secret_key',
    'load_monitoring',
    'load_log_queries',
    'load_throttling',
    'load_default_cache_exp',
    'load_middlewares',
    'load_user_model',
    'load_authentication_class',
    'load_jwt_config',
    'collect_all_models',
    'load_urls',
    'load_panel_urls',
)


def load_configs_file(_configs, /):
    """Read the config file and put it as dict in self.configs"""
    if _configs is None:
        try:
            _configs = import_module('core.configs')
        except ModuleNotFoundError:
            raise _exception_handler(field='core/configs.py', error='Not Found')
    return _configs.__dict__


def load_secret_key(configs: dict, /) -> bytes | None:
    if secret_key := configs.get('SECRET_KEY'):
        return secret_key.encode()
    return secret_key


def load_monitoring(configs: dict, /) -> bool:
    return configs.get('MONITORING', config['monitoring'])


def load_log_queries(configs: dict, /) -> bool:
    return configs.get('LOG_QUERIES', config['log_queries'])


def load_throttling(configs: dict, /) -> Throttling | None:
    return configs.get('THROTTLING', config['throttling'])


def load_default_cache_exp(configs: dict, /) -> timedelta | None:
    return configs.get('DEFAULT_CACHE_EXP', config['default_cache_exp'])


def load_middlewares(configs: dict, /) -> list:
    """Collect The Middlewares & Set db_engine If One Of Middlewares Was For DB"""
    middlewares = list()

    for path, data in configs.get('MIDDLEWARES', []):
        if path.find('panther.middlewares.db.DatabaseMiddleware') != -1:
            config['db_engine'] = data['url'].split(':')[0]

        Middleware = import_class(path)  # noqa: N806
        if not issubclass(Middleware, BaseMiddleware):
            raise _exception_handler(field='MIDDLEWARES', error='is not a sub class of BaseMiddleware')

        middlewares.append(Middleware(**data))  # noqa: Py Argument List
    return middlewares


def load_user_model(configs: dict, /) -> ModelMetaclass:
    return import_class(configs.get('USER_MODEL', 'panther.db.models.BaseUser'))


def load_authentication_class(configs: dict, /) -> ModelMetaclass | None:
    return configs.get('AUTHENTICATION') and import_class(configs['AUTHENTICATION'])


def load_jwt_config(configs: dict, /) -> JWTConfig:
    """Only Collect JWT Config If Authentication Is JWTAuthentication"""
    if getattr(config['authentication'], '__name__', None) == 'JWTAuthentication':
        user_config = configs.get('JWTConfig')
        return JWTConfig(**user_config) if user_config else JWTConfig(key=config['secret_key'].decode())


def collect_all_models():
    """Collecting all models for panel APIs"""
    from panther.db.models import Model
    collected_models = list()

    for root, _, files in os.walk(config['base_dir']):
        # Traverse through each directory
        for f in files:
            # Traverse through each file of directory
            if f == 'models.py':
                slash = '\\' if platform.system() == 'Windows' else '/'

                # If the file was "models.py" read it
                file_path = f'{root}{slash}models.py'
                with open(file_path) as file:
                    # Parse the file with ast
                    node = ast.parse(file.read())
                    for n in node.body:
                        # Find classes in each element of files' body
                        if type(n) is ast.ClassDef and n.bases:
                            class_path = file_path \
                                .removesuffix(f'{slash}models.py') \
                                .removeprefix(f'{config["base_dir"]}{slash}') \
                                .replace(slash, '.')
                            # We don't need to import the package classes
                            if class_path.find('site-packages') == -1:
                                # Import the class to check his parents and siblings
                                klass = import_class(f'{class_path}.models.{n.name}')

                                collected_models.extend([
                                    {
                                        'name': n.name,
                                        'path': file_path,
                                        'class': klass,
                                        'app': class_path.split('.'),
                                    }
                                    for parent in klass.__mro__ if parent is Model
                                ])
    return collected_models


def load_urls(configs: dict, /, urls: dict | None) -> dict:
    if isinstance(urls, dict):
        return urls

    try:
        urls = import_class(configs.get('URLs'))
    except ModuleNotFoundError as e:
        raise _exception_handler(field='URLs', error=e)

    if not isinstance(urls, dict):
        raise _exception_handler(field='URLs', error='should point to a dict')

    collected_urls = flatten_urls(urls)
    return finalize_urls(collected_urls)


def load_panel_urls() -> dict:
    from panther.panel.urls import urls

    return finalize_urls(flatten_urls(urls))


def _exception_handler(field: str, error: str | Exception):
    cli_error(message=f"[Invalid '{field}'] {error}")
    return TypeError



from datetime import datetime

from panther import version
from panther.utils import generate_secret_key

apis_py = """from datetime import datetime

from app.throttling import InfoThrottling

from panther import status, version
from panther.app import API
from panther.request import Request
from panther.response import Response


@API()
async def hello_world_api():
    return {'detail': 'Hello World'}


@API(cache=True, throttling=InfoThrottling)
async def info_api(request: Request):
    data = {
        'panther_version': version(),
        'datetime_now': datetime.now().isoformat(),
        'user_agent': request.headers.user_agent,
    }
    return Response(data=data, status_code=status.HTTP_202_ACCEPTED)
"""

models_py = """from panther.db import Model
"""

serializers_py = """from pydantic import BaseModel
"""

throttling_py = """from datetime import timedelta

from panther.throttling import Throttling

InfoThrottling = Throttling(rate=5, duration=timedelta(minutes=1))
"""

app_urls_py = """from app.apis import hello_world_api, info_api

urls = {
    '/': hello_world_api,
    'info/': info_api,
}
"""

configs_py = """\"""
{PROJECT_NAME} Project (Generated by Panther on %s)
\"""

from datetime import timedelta
from pathlib import Path

from panther.throttling import Throttling
from panther.utils import load_env

BASE_DIR = Path(__name__).resolve().parent
env = load_env(BASE_DIR / '.env')

SECRET_KEY = env['SECRET_KEY']

# More Info: Https://PantherPy.GitHub.io/middlewares/
MIDDLEWARES = [
    ('panther.middlewares.db.DatabaseMiddleware', {'url': f'pantherdb://{BASE_DIR}/database.pdb'}),
]

# More Info: https://PantherPy.GitHub.io/configs/#user_model
USER_MODEL = 'panther.db.models.BaseUser'

# More Info: https://PantherPy.GitHub.io/authentications/
AUTHENTICATION = 'panther.authentications.JWTAuthentication'

# More Info: https://PantherPy.GitHub.io/monitoring/
MONITORING = True

# More Info: https://PantherPy.GitHub.io/log_queries/
LOG_QUERIES = True

# More Info: https://PantherPy.GitHub.io/throttling/
THROTTLING = Throttling(rate=60, duration=timedelta(minutes=1))

# More Info: https://PantherPy.GitHub.io/urls/
URLs = 'core.urls.url_routing'
""" % datetime.now().date().isoformat()

env = """
SECRET_KEY = '%s'
""" % generate_secret_key()

main_py = """from panther import Panther

app = Panther(__name__)
"""

urls_py = """from app.urls import urls as app_urls

url_routing = {
    '/': app_urls,
}
"""

git_ignore = """__pycache__/
.venv/
.idea/
logs/

.env
*.pdb
"""

requirements = """panther==%s
""" % version()

Template = {
    'app': {
        'apis.py': apis_py,
        'models.py': models_py,
        'serializers.py': serializers_py,
        'throttling.py': throttling_py,
        'urls.py': app_urls_py,
    },
    'core': {
        'configs.py': configs_py,
        'urls.py': urls_py,
    },
    'main.py': main_py,
    '.env': env,
    '.gitignore': git_ignore,
    'requirements.txt': requirements,
}
